import numba
import numpy as np


def compute_DtD(D, n_channels=None):
    """Compute the DtD matrix
    """
    if D.ndim == 2:
        return _compute_DtD_uv(D, n_channels)

    return _compute_DtD_D(D)


@numba.jit((numba.float64[:, :], numba.int64), nopython=True, cache=True)
def _compute_DtD_uv(uv, n_channels):  # pragma: no cover
    # TODO: benchmark the cross correlate function of numpy
    n_atoms, n_times_atom = uv.shape
    n_times_atom -= n_channels

    u = uv[:, :n_channels]

    DtD = np.zeros(shape=(n_atoms, n_atoms, 2 * n_times_atom - 1))
    t0 = n_times_atom - 1
    for k0 in range(n_atoms):
        for k in range(n_atoms):
            for t in range(n_times_atom):
                if t == 0:
                    DtD[k0, k, t0] = np.dot(uv[k0, n_channels:],
                                            uv[k, n_channels:])
                else:
                    DtD[k0, k, t0 + t] = np.dot(uv[k0, n_channels:-t],
                                                uv[k, n_channels + t:])
                    DtD[k0, k, t0 - t] = np.dot(uv[k0, n_channels + t:],
                                                uv[k, n_channels:-t])
    DtD *= np.dot(u, u.T).reshape(n_atoms, n_atoms, 1)
    return DtD


@numba.jit((numba.float64[:, :, :],), nopython=True, cache=True)
def _compute_DtD_D(D):  # pragma: no cover
    # TODO: benchmark the cross correlate function of numpy
    n_atoms, n_channels, n_times_atom = D.shape

    DtD = np.zeros(shape=(n_atoms, n_atoms, 2 * n_times_atom - 1))
    t0 = n_times_atom - 1
    for k0 in range(n_atoms):
        for k in range(n_atoms):
            for t in range(n_times_atom):
                if t == 0:
                    DtD[k0, k, t0] = np.dot(D[k0].ravel(), D[k].ravel())
                else:
                    DtD[k0, k, t0 + t] = np.dot(D[k0, :, :-t].ravel(),
                                                D[k, :, t:].ravel())
                    DtD[k0, k, t0 - t] = np.dot(D[k0, :, t:].ravel(),
                                                D[k, :, :-t].ravel())
    return DtD


@numba.jit((numba.float64[:, :, :], numba.int64), nopython=True, cache=True)
def compute_ztz(z, n_times_atom):  # pragma: no cover
    """
    ztz.shape = n_atoms, n_atoms, 2 * n_times_atom - 1
    z.shape = n_trials, n_atoms, n_times - n_times_atom + 1)
    """
    # TODO: benchmark the cross correlate function of numpy
    n_trials, n_atoms, n_times_valid = z.shape

    ztz = np.zeros(shape=(n_atoms, n_atoms, 2 * n_times_atom - 1))
    t0 = n_times_atom - 1
    for i in range(n_trials):
        for k0 in range(n_atoms):
            for k in range(n_atoms):
                for t in range(n_times_atom):
                    if t == 0:
                        ztz[k0, k, t0] += (z[i, k0] * z[i, k]).sum()
                    else:
                        ztz[k0, k, t0 + t] += (
                            z[i, k0, :-t] * z[i, k, t:]).sum()
                        ztz[k0, k, t0 - t] += (
                            z[i, k0, t:] * z[i, k, :-t]).sum()
    return ztz


def compute_ztX(z, X):
    """
    z.shape = n_trials, n_atoms, n_times - n_times_atom + 1)
    X.shape = n_trials, n_channels, n_times
    ztX.shape = n_atoms, n_channels, n_times_atom
    """
    n_trials, n_atoms, n_times_valid = z.shape
    _, n_channels, n_times = X.shape
    n_times_atom = n_times - n_times_valid + 1

    ztX = np.zeros((n_atoms, n_channels, n_times_atom))
    for n, k, t in zip(*z.nonzero()):
        ztX[k, :, :] += z[n, k, t] * X[n, :, t:t + n_times_atom]

    return ztX



import time
import numpy as np
from scipy import sparse

from .lil import is_lil
from .. import cython_code
from ..loss_and_gradient import gradient_zi
from .convolution import _choose_convolve_multi


def _coordinate_descent_idx(Xi, D, constants, reg, z0=None, max_iter=1000,
                            tol=1e-1, strategy='greedy', n_seg='auto',
                            freeze_support=False, debug=False, timing=False,
                            name="CD", verbose=0):
    """Compute the coding signal associated to Xi with coordinate descent.

    Parameters
    ----------
    Xi : array, shape (n_channels, n_times)
        The signal to encode.
    D : array
        The atoms. Can either be full rank with shape shape
        (n_atoms, n_channels, n_times_atom) or rank 1 with
        shape shape (n_atoms, n_channels + n_times_atom)
    constants : dict
        Constants containing DtD to speedup computation
    z0 : array, shape (n_atoms, n_times_valid)
        Initial estimate of the coding signal, to warm t_start the algorithm.
    tol : float
        Tolerance for the stopping criterion of the algorithm
    max_iter : int
        Maximal number of iterations run by the algorithm
    strategy : str in {'greedy' | 'random'}
        Strategy to select the updated coordinate in the CD algorithm.
    n_seg : int or 'auto'
        Number of segments used to divide the coding signal. The updates are
        performed successively on each of these segments.
    freeze_support : boolean
        If set to True, only update the coefficient that are non-zero in z0.
    debug : boolean
        Activate extra check in the algorithm to assert that we have
        implemented the correct algorithm.
    """
    if timing:
        t_start = time.time()
    n_channels, n_times = Xi.shape
    if D.ndim == 2:
        n_atoms, n_times_atom = D.shape
        n_times_atom -= n_channels
    else:
        n_atoms, n_channels, n_times_atom = D.shape
    n_times_valid = n_times - n_times_atom + 1
    t0 = n_times_atom - 1

    if z0 is None:
        z_hat = np.zeros((n_atoms, n_times_valid))
    else:
        z_hat = z0.copy()

    if n_seg == 'auto':
        if strategy == 'greedy':
            n_seg = max(n_times_valid // (2 * n_times_atom), 1)
        elif strategy in ('random', 'cyclic'):
            n_seg = 1
            n_coordinates = n_times_valid * n_atoms

    max_iter *= n_seg
    n_times_seg = n_times_valid // n_seg + 1

    def objective(zi):
        Dzi = _choose_convolve_multi(zi, D=D, n_channels=n_channels)
        Dzi -= Xi
        func = 0.5 * np.dot(Dzi.ravel(), Dzi.ravel())
        func += reg * zi.sum()
        return func

    DtD = constants["DtD"]
    norm_Dk = np.array([DtD[k, k, t0] for k in range(n_atoms)])[:, None]

    if timing:
        times = [time.time() - t_start]
        pobj = [objective(z_hat)]
        t_start = time.time()

    beta, dz_opt = _init_beta(Xi, z_hat, D, constants, reg, norm_Dk,
                              tol, use_sparse_dz=False)

    # If we freeze the support, we put dz_opt to zero outside the support of z0
    if freeze_support:
        if is_lil(z0):
            mask = z0 != 0
            mask = mask.toarray()
            mask = ~mask
        else:
            mask = z0 == 0
        dz_opt[mask] = 0

    accumulator = n_seg
    active_segs = np.array([True] * n_seg)
    i_seg = 0
    seg_bounds = [0, n_times_seg]
    t0, k0 = -1, 0
    for ii in range(int(max_iter)):
        k0, t0, dz = _select_coordinate(strategy, dz_opt, active_segs[i_seg],
                                        n_atoms, n_times_valid, n_times_seg,
                                        seg_bounds)
        if strategy in ['random', 'cyclic']:
            # accumulate on all coordinates from the stopping criterion
            if ii % n_coordinates == 0:
                accumulator = 0
            accumulator += abs(dz)

        # Update the selected coordinate and beta, only if the update is
        # greater than the convergence tolerance.
        if abs(dz) > tol:
            # update the selected coordinate
            z_hat[k0, t0] += dz

            # update beta
            beta, dz_opt, accumulator, active_segs = _update_beta(
                beta, dz_opt, accumulator, active_segs, z_hat, DtD, norm_Dk,
                dz, k0, t0, reg, tol, seg_bounds, i_seg, n_times_atom, z0,
                freeze_support, debug)

        elif active_segs[i_seg]:
            accumulator -= 1
            active_segs[i_seg] = False

        if timing and (ii % max(100, n_seg // 100) == 0):
            times.append(time.time() - t_start)
            pobj.append(objective(z_hat))
            t_start = time.time()

        # check stopping criterion
        if strategy == 'greedy':
            if accumulator == 0:
                if verbose > 10:
                    print('[{}] {} iterations'.format(name, ii + 1))
                break
        else:
            # only check at the last coordinate
            if (ii + 1) % n_coordinates == 0 and accumulator <= tol:
                if verbose > 10:
                    print('[{}] {} iterations'.format(name, ii + 1))
                break

        # increment to next segment
        i_seg += 1
        seg_bounds[0] += n_times_seg
        seg_bounds[1] += n_times_seg

        if seg_bounds[0] >= n_times_valid:
            # reset to first segment
            i_seg = 0
            seg_bounds = [0, n_times_seg]

    else:
        if verbose > 10:
            print('[{}] did not converge'.format(name))

    if timing:
        return z_hat, pobj, times
    return z_hat


def _init_beta(Xi, z_hat, D, constants, reg, norm_Dk, tol,
               use_sparse_dz=False):
    # Init beta with -DtX
    beta = gradient_zi(Xi, z_hat, D=D, reg=None, loss='l2',
                       return_func=False, constants=constants)
    if is_lil(z_hat):
        cython_code._assert_cython()
        beta = cython_code.subtract_zhat_to_beta(beta, z_hat, norm_Dk[:, 0])
    else:
        for k, t in zip(*z_hat.nonzero()):
            beta[k, t] -= z_hat[k, t] * norm_Dk[k]  # np.sum(DtD[k, k, t0])

    if is_lil(z_hat):
        n_times_valid = beta.shape[1]
        dz_opt = np.zeros(beta.shape)
        cython_code._assert_cython()
        cython_code.update_dz_opt(z_hat, beta, dz_opt, norm_Dk[:, 0], reg,
                                  t_start=0, t_end=n_times_valid)
    else:
        dz_opt = np.maximum(-beta - reg, 0) / norm_Dk - z_hat

    if use_sparse_dz:
        dz_opt[abs(dz_opt) < tol] = 0
        dz_opt = sparse.lil_matrix(dz_opt)

    return beta, dz_opt


def _update_beta(beta, dz_opt, accumulator, active_segs, z_hat, DtD, norm_Dk,
                 dz, k0, t0, reg, tol, seg_bounds, i_seg, n_times_atom, z0,
                 freeze_support, debug):
    n_atoms, n_times_valid = beta.shape

    # define the bounds for the beta update
    t_start_up = max(0, t0 - n_times_atom + 1)
    t_end_up = min(t0 + n_times_atom, n_times_valid)

    # update beta
    beta_i0 = beta[k0, t0]
    ll = t_end_up - t_start_up
    offset = max(0, n_times_atom - t0 - 1)
    beta[:, t_start_up:t_end_up] += DtD[:, k0, offset:offset + ll] * dz
    beta[k0, t0] = beta_i0

    # update dz_opt
    if is_lil(z_hat):
        cython_code._assert_cython()
        cython_code.update_dz_opt(
            z_hat, beta, dz_opt, norm_Dk[:, 0], reg, t_start_up, t_end_up)
    else:
        tmp = np.maximum(-beta[:, t_start_up:t_end_up] - reg, 0) / norm_Dk
        dz_opt[:, t_start_up:t_end_up] = tmp - z_hat[:, t_start_up:t_end_up]
    dz_opt[k0, t0] = 0

    # reunable greedy updates in the segments immediately before or after
    # if beta was update outside the segment
    t_start_seg, t_end_seg = seg_bounds
    if t_start_up < t_start_seg and not active_segs[i_seg - 1]:
        accumulator += 1
        active_segs[i_seg - 1] = True
    if t_end_up > t_end_seg and not active_segs[i_seg + 1]:
        accumulator += 1
        active_segs[i_seg + 1] = True

    # If we freeze the support, we put dz_opt to zero outside the support of z0
    if freeze_support:
        if is_lil(z0):
            mask = z0[:, t_start_up:t_end_up] != 0
            mask = mask.toarray()
            mask = ~mask
        else:
            mask = z0[:, t_start_up:t_end_up] == 0
        dz_opt[:, t_start_up:t_end_up][mask] = 0

        if debug:
            # Check that we do not changed the support while updating beta
            nnz_z0 = list(zip(*z0[:, t_start_up:t_end_up].nonzero()))
            nnz_dz = list(zip(*dz_opt[:, t_start_up:t_end_up].nonzero()))
            assert all([nnz in nnz_z0 for nnz in nnz_dz])

    return beta, dz_opt, accumulator, active_segs


def _select_coordinate(strategy, dz_opt, active_seg, n_atoms, n_times_valid,
                       n_times_seg, seg_bounds):
    # Pick a coordinate to update
    if strategy == 'random':
        k0 = np.random.randint(n_atoms)
        t0 = np.random.randint(n_times_valid)
        dz = dz_opt[k0, t0]

    elif strategy == 'cyclic':
        t0 += 1
        if t0 >= n_times_valid:
            t0 = 0
            k0 += 1
            if k0 >= n_atoms:
                k0 = 0
        dz = dz_opt[k0, t0]

    elif strategy == 'greedy':
        # if dZs[i_seg] > tol:
        t_start_seg, t_end_seg = seg_bounds
        if active_seg:
            i0 = abs(dz_opt[:, t_start_seg:t_end_seg]).argmax()
            n_times_current = min(n_times_seg, n_times_valid - t_start_seg)
            k0, t0 = np.unravel_index(i0, (n_atoms, n_times_current))
            t0 += t_start_seg
            dz = dz_opt[k0, t0]
        else:
            k0, t0, dz = None, None, 0
    else:
        raise ValueError("'The coordinate selection method should be in "
                         "{'greedy' | 'random' | 'cyclic'}. Got '%s'."
                         % (strategy, ))
    return k0, t0, dz



"""
==========================================================
Alpha CSC on empirical time-series with strong artifacts
==========================================================

This example illustrates how to learn univariate atoms on a univariate
time-serie affected by strong artifacts. The data is a single LFP channel
recorded on a rodent's striatum [1]_. Interestingly in this time-serie, the
high frequency oscillations around 80 Hz are modulated in amplitude by the
low-frequency oscillation around 3 Hz, a phenomenon known as cross-frequency
coupling (CFC).

The convolutional sparse coding (CSC) model is able to learn the prototypical
waveforms of the signal, on which we can clearly see the CFC. However, when the
CSC model is fitted on a data section with strong artifacts, the learned atoms
do not show the expected CFC waveforms. To solve this problem, another model
can be used, called alpha-CSC [2]_, which is less affected by strong artifacts
in the data.

.. [1] G. Dallérac, M. Graupner, J. Knippenberg, R. C. R. Martinez,
    T. F. Tavares, L. Tallot, N. El Massioui, A. Verschueren, S. Höhn,
    J.B. Bertolus, et al. Updating temporal expectancy of an aversive event
    engages striatal plasticity under amygdala control.
    Nature Communications, 8:13920, 2017

.. [2] Jas, M., Dupré La Tour, T., Şimşekli, U., & Gramfort, A. (2017).
    `Learning the Morphology of Brain Signals Using Alpha-Stable Convolutional
    Sparse Coding
    <https://papers.nips.cc/paper/6710-learning-the-morphology-of-brain-signals-using-alpha-stable-convolutional-sparse-coding.pdf>`_.
    Advances in Neural Information Processing Systems (NIPS), pages 1099--1108.
"""

# Authors: Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@telecom-paristech.fr>
#
# License: BSD (3-clause)

###############################################################################
# Let us first load the data sample.

import mne
import numpy as np
import matplotlib.pyplot as plt

# sample frequency
sfreq = 350.

# We load the signal. It is an LFP channel recorded on a rodent's striatum.
data = np.load('../rodent_striatum.npy')
print(data.shape)

###############################################################################
# Now let us take a closer look, plotting the 100 first seconds of the signal.

start, stop = [0, 100]  # in seconds
start, stop = int(start * sfreq), int(stop * sfreq)
time = np.arange(start, stop) / sfreq
plt.plot(time, data[0, start:stop])
plt.show()

###############################################################################
# As we can see, the data contains severe artifacts. We will thus compare three
# approaches to tackle these artifacts:
#   - First, we will fit a CSC model on a section not affected by artifacts.
#   - Then, we will fit a CSC model on a section affected by artifacts.
#   - Finally, we will fit an alpha-CSC model on a section affected by
#     artifacts.

# Define a clean data section.
start, stop = [100, 600]  # in seconds
start, stop = int(start * sfreq), int(stop * sfreq)
data_clean = data[:, start:stop].copy()

# Define a dirty data section (same length).
start, stop = [0, 500]  # in seconds
start, stop = int(start * sfreq), int(stop * sfreq)
data_dirty = data[:, start:stop].copy()

# We also remove the slow drift, which accounts for a lot of variance.
data_clean = mne.filter.filter_data(data_clean, sfreq, 1, None)
data_dirty = mne.filter.filter_data(data_dirty, sfreq, 1, None)

# To make the most of parallel computing, we split the data into trials.
data_clean = data_clean.reshape(50, -1)
data_dirty = data_dirty.reshape(50, -1)

# We scale the data, since parameter alpha is scale dependant.
scale = data_clean.std()
data_clean /= scale
data_dirty /= scale

###############################################################################
# This sample contains CFC between 3 Hz and 80 Hz. This phenomenon can be
# described with a comodulogram, computed for instance with the `pactools
# <http://pactools.github.io/>`_ Python library.

from pactools import Comodulogram

comod = Comodulogram(fs=sfreq, low_fq_range=np.arange(0.2, 10.2, 0.2),
                     low_fq_width=2., method='duprelatour')
comod.fit(data_clean)
comod.plot()
plt.show()

###############################################################################
# Here we define the plotting function which display the learned atoms.


def plot_atoms(d_hat):
    n_atoms, n_times_atom = d_hat.shape
    n_columns = min(6, n_atoms)
    n_rows = int(np.ceil(n_atoms // n_columns))
    figsize = (4 * n_columns, 3 * n_rows)
    fig, axes = plt.subplots(n_rows, n_columns, figsize=figsize, sharey=True)
    axes = axes.ravel()

    # Plot the temporal pattern of the atom
    for kk in range(n_atoms):
        ax = axes[kk]
        time = np.arange(n_times_atom) / sfreq
        ax.plot(time, d_hat[kk], color='C%d' % kk)
        ax.set_xlim(0, n_times_atom / sfreq)
        ax.set(xlabel='Time (sec)', title="Temporal pattern %d" % kk)
        ax.grid(True)

    fig.tight_layout()
    plt.show()


###############################################################################
# Then we define the common parameters of the different models.

common_params = dict(
    n_atoms=3,
    n_times_atom=int(sfreq * 1.0),  # 1000. ms
    reg=3.,
    solver_z='l-bfgs',
    solver_z_kwargs=dict(factr=1e9),
    solver_d_kwargs=dict(factr=1e2),
    random_state=42,
    n_jobs=5,
    verbose=1)

# number of iterations
n_iter = 10

# Parameter of the alpha-stable distribution for the alpha-CSC model.
# 0 < alpha < 2
# A value of 2 would correspond to the Gaussian noise model, as in vanilla CSC.
alpha = 1.2


###############################################################################
# First, we fit a CSC model on the clean data. Interestingly, we obtain
# prototypical waveforms of the signal on which we can clearly see the CFC.

from alphacsc import learn_d_z, learn_d_z_weighted

X = data_clean

_, _, d_hat, z_hat, _ = learn_d_z(X, n_iter=n_iter, **common_params)

plot_atoms(d_hat)

###############################################################################
# Then, if we fit a CSC model on the dirty data, the model is strongly affected
# by the artifacts, and we cannot see CFC anymore in the temporal waveforms.

X = data_dirty

_, _, d_hat, z_hat, _ = learn_d_z(X, n_iter=n_iter, **common_params)

plot_atoms(d_hat)

###############################################################################
# Finally, If we fit an alpha-CSC model on the dirty data, the model is less
# affected by the artifacts, and we are able to see CFC in the temporal
# waveforms.

X = data_dirty

d_hat, z_hat, tau = learn_d_z_weighted(
    X, n_iter_optim=n_iter, n_iter_global=3, n_iter_mcmc=300,
    n_burnin_mcmc=100, alpha=alpha, **common_params)

plot_atoms(d_hat)



# Authors: Mainak Jas <mainak.jas@telecom-paristech.fr>
#          Tom Dupre La Tour <tom.duprelatour@telecom-paristech.fr>
#          Umut Simsekli <umut.simsekli@telecom-paristech.fr>
#          Alexandre Gramfort <alexandre.gramfort@inria.fr>
#          Thomas Moreau <thomas.moreau@inria.fr>
import itertools

import numpy as np
import matplotlib.pyplot as plt

from sklearn.cluster import MiniBatchKMeans
from sklearn.decomposition import PCA
from sklearn.manifold import TSNE

from .viz.callback import COLORS
from .utils import check_random_state
from .other.kmc2 import custom_distances
from .update_d_multi import prox_uv, prox_d
from .utils.dictionary import get_uv, get_D, _patch_reconstruction_error
from .utils.dictionary import tukey_window

ried = custom_distances.roll_invariant_euclidean_distances
tied = custom_distances.translation_invariant_euclidean_distances


def init_dictionary(X, n_atoms, n_times_atom, uv_constraint='separate',
                    rank1=True, window=False, D_init=None,
                    D_init_params=dict(), random_state=None):
    """Return an initial dictionary for the signals X

    Parameter
    ---------
    X: array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    uv_constraint : str in {'joint' | 'separate'}
        The kind of norm constraint on the atoms:
        If 'joint', the constraint is norm_2([u, v]) <= 1
        If 'separate', the constraint is norm_2(u) <= 1 and norm_2(v) <= 1
    rank1 : boolean
        If set to True, use a rank 1 dictionary.
    window : boolean
        If True, multiply the atoms with a temporal Tukey window.
    D_init : array or {'kmeans' | 'ssa' | 'chunk' | 'random'}
        The initialization scheme for the dictionary or the initial
        atoms. The shape should match the required dictionary shape, ie if
        rank1 is TRue, (n_atoms, n_channels + n_times_atom) and else
        (n_atoms, n_channels, n_times_atom)
    D_init_params : dict
        Dictionnary of parameters for the kmeans init method.
    random_state : int | None
        The random state.

    Return
    ------
    D : array shape (n_atoms, n_channels + n_times_atom) or
              shape (n_atoms, n_channels, n_times_atom)
        The initial atoms to learn from the data.
    """
    n_trials, n_channels, n_times = X.shape
    rng = check_random_state(random_state)

    D_shape = (n_atoms, n_channels, n_times_atom)
    if rank1:
        D_shape = (n_atoms, n_channels + n_times_atom)

    if isinstance(D_init, np.ndarray):
        D_hat = D_init.copy()
        assert D_hat.shape == D_shape

    elif D_init is None or D_init == "random":
        D_hat = rng.randn(*D_shape)

    elif D_init == 'chunk':
        D_hat = np.zeros((n_atoms, n_channels, n_times_atom))
        for i_atom in range(n_atoms):
            i_trial = rng.randint(n_trials)
            t0 = rng.randint(n_times - n_times_atom)
            D_hat[i_atom] = X[i_trial, :, t0:t0 + n_times_atom]
        if rank1:
            D_hat = get_uv(D_hat)

    elif D_init == "kmeans":
        D_hat = kmeans_init(X, n_atoms, n_times_atom, random_state=rng,
                            **D_init_params)
        if not rank1:
            D_hat = get_D(D_hat, n_channels)

    elif D_init == "ssa":
        u_hat = rng.randn(n_atoms, n_channels)
        v_hat = ssa_init(X, n_atoms, n_times_atom, random_state=rng)
        D_hat = np.c_[u_hat, v_hat]
        if not rank1:
            D_hat = get_D(D_hat, n_channels)

    elif D_init == 'greedy':
        raise NotImplementedError()

    else:
        raise NotImplementedError('It is not possible to initialize uv with'
                                  ' parameter {}.'.format(D_init))

    if window and not isinstance(D_init, np.ndarray):
        if rank1:
            D_hat[:, n_channels:] *= tukey_window(n_times_atom)[None, :]
        else:
            D_hat = D_hat * tukey_window(n_times_atom)[None, None, :]

    if rank1:
        D_hat = prox_uv(D_hat, uv_constraint=uv_constraint,
                        n_channels=n_channels)
    else:
        D_hat = prox_d(D_hat)
    return D_hat


def kmeans_init(X, n_atoms, n_times_atom, max_iter=0, random_state=None,
                non_uniform=True, distances='euclidean', tsne=False):
    """Return an initial temporal dictionary for the signals X

    Parameter
    ---------
    X : array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    max_iter : int
        Number of iteration of kmeans algorithm
    random_state : int | None
        The random state.
    non_uniform : boolean
        If True, the kmc2 init uses the norm of each data chunk.
    distances : str in {'euclidean', 'roll_inv', 'trans_inv'}
        Distance kind.

    Return
    ------
    uv: array shape (n_atoms, n_channels + n_times_atom)
        The initial atoms to learn from the data.
    """
    rng = check_random_state(random_state)

    n_trials, n_channels, n_times = X.shape
    X_original = X
    if distances != 'euclidean':
        # Only take the strongest channels, otherwise X is too big
        n_strong_channels = 1
        strongest_channels = np.argsort(X.std(axis=2).mean(axis=0))
        X = X[:, strongest_channels[-n_strong_channels:], :]

    X = X.reshape(-1, X.shape[-1])
    n_trials, n_times = X.shape

    # Time step between two windows
    step = max(1, n_times_atom // 3)

    # embed all windows of length n_times_atom in X
    X_embed = np.concatenate(
        [_embed(Xi, n_times_atom).T[::step, :] for Xi in X])
    X_embed = np.atleast_2d(X_embed)

    if non_uniform:
        weights = np.linalg.norm(X_embed, axis=1)
    else:
        weights = None

    # init the kmeans centers with KMC2
    try:
        from alphacsc.other.kmc2 import kmc2
        seeding, indices = kmc2.kmc2(X_embed, k=n_atoms, weights=weights,
                                     random_state=rng, distances=distances)
    except ImportError:
        if max_iter == 0:
            raise ImportError("Could not import alphacsc.other.kmc2. This "
                              "breaks the logic for the D_init='kmeans'. It "
                              "should not be used with max_iter=0 in "
                              "D_init_params.")
        # Default to random init for non-euclidean distances and to "kmeans++"
        # in the case of K-means.
        indices = rng.choice(len(X_embed), size=n_atoms, replace=False)
        seeding = "kmeans++"

    # perform the kmeans, or use the seeding if max_iter == 0
    if max_iter == 0:
        v_init = seeding
        labels = None
        distance_metric = 'euclidean'

    elif distances != 'euclidean':
        if distances == 'trans_inv':
            distance_metric = tied
        elif distances == 'roll_inv':
            distance_metric = ried
        else:
            raise ValueError('Unknown distance "%s".' % (distances, ))

        try:
            from .other.k_medoids import KMedoids
        except ImportError:
            raise ImportError("Could not import multics.other.k_medoid, make "
                              "sure to compile it to be able to initialize "
                              "the dictionary with k-means and a non-euclidean"
                              " distance.")
        model = KMedoids(n_clusters=n_atoms, init=np.int_(indices),
                         max_iter=max_iter, distance_metric=distance_metric,
                         random_state=rng).fit(X_embed)
        indices = model.medoid_idxs_
        labels = model.labels_

    else:
        distance_metric = 'euclidean'
        model = MiniBatchKMeans(n_clusters=n_atoms, init=seeding, n_init=1,
                                max_iter=max_iter, random_state=rng
                                ).fit(X_embed)
        v_init = model.cluster_centers_
        u_init = rng.randn(n_atoms, n_channels)
        D_init = np.c_[u_init, v_init]
        labels = model.labels_

    if tsne:
        if distances == 'euclidean':
            X_embed = X_embed[::100]
            if labels is not None:
                labels = labels[::100]
        plot_tsne(X_embed, v_init, labels=labels, metric=distance_metric,
                  random_state=rng)

    if not (distances == 'euclidean' and max_iter > 0):
        indices = np.array(indices)
        n_window = X_embed.shape[0] // n_trials
        medoid_i = (indices // n_window) // n_channels
        medoid_t = (indices % n_window) * step
        D = np.array([X_original[i, :, t:t + n_times_atom]
                      for i, t in zip(medoid_i, medoid_t)])
        D_init = get_uv(D)

    return D_init


def plot_tsne(X_embed, X_centers, labels=None, metric='euclidean',
              random_state=None):

    tsne = TSNE(n_components=2, random_state=random_state, perplexity=5,
                metric=metric, verbose=2)
    pca = PCA(n_components=min(10, X_embed.shape[1]))
    X = np.concatenate([X_embed, X_centers])
    n_centers = X_centers.shape[0]
    X_pca = pca.fit_transform(X)
    X_tsne = tsne.fit_transform(X_pca)

    if labels is not None:
        labels = np.r_[labels, np.arange(n_centers)]
        colors = [c for c, l in zip(itertools.cycle(COLORS),
                                    np.unique(labels))]
        colors = np.array(colors)[labels]
    else:
        colors = None

    plt.figure('tsne')
    cc = colors[:-n_centers] if colors is not None else None
    plt.scatter(X_tsne[:-n_centers, 0], X_tsne[:-n_centers, 1], c=cc,
                marker='.')
    cc = colors[-n_centers:] if colors is not None else None
    plt.scatter(X_tsne[-n_centers:, 0], X_tsne[-n_centers:, 1], c=cc,
                marker='*', s=4)


def ssa_init(X, n_atoms, n_times_atom, random_state=None):
    """Return an initial temporal dictionary for the signals X

    Parameter
    ---------
    X: array, shape (n_trials, n_channels, n_times)
        The data on which to perform CSC.
    n_atoms : int
        The number of atoms to learn.
    n_times_atom : int
        The support of the atom.
    random_state : int | None
        The random state.

    Return
    ------
    uv: array shape (n_atoms, n_channels + n_times_atom)
        The initial atoms to learn from the data.
    """
    # Only take the strongest channel, otherwise X is too big
    strongest_channel = np.argmax(X.std(axis=2).mean(axis=0))
    X_strong = X[:, strongest_channel, :]

    # Time step between two windows
    step = 1

    # embed all the windows of length n_times_atom in X_strong
    X_embed = np.concatenate(
        [_embed(Xi, n_times_atom).T[::step, :] for Xi in X_strong])
    X_embed = np.atleast_2d(X_embed)

    model = PCA(n_components=n_atoms, random_state=random_state).fit(X_embed)
    v_init = model.components_

    return v_init


def _embed(x, dim, lag=1):
    """Create an embedding of array given a resulting dimension and lag.
    """
    x = x.copy()
    X = np.lib.stride_tricks.as_strided(x, (len(x) - dim * lag + lag, dim),
                                        (x.strides[0], x.strides[0] * lag))
    return X.T


def get_max_error_dict(X, z, D, uv_constraint='separate', window=False):
    """Get the maximal reconstruction error patch from the data as a new atom

    This idea is used for instance in [Yellin2017]

    Parameters
    ----------
    X: array, shape (n_trials, n_channels, n_times)
        Signals encoded in the CSC.
    z: array, shape (n_atoms, n_trials, n_times_valid)
        Current estimate of the coding signals.
    D: array, shape (n_atoms, n_channels + n_times_atom)
        Current estimate of the rank1 multivariate dictionary.
    uv_constraint : str in {'joint' | 'separate'}
        The kind of norm constraint on the atoms:
        If 'joint', the constraint is norm_2([u, v]) <= 1
        If 'separate', the constraint is norm_2(u) <= 1 and norm_2(v) <= 1
    window : boolean
        If True, multiply the atoms with a temporal Tukey window.

    Return
    ------
    uvk: array, shape (n_channels + n_times_atom,)
        New atom for the dictionary, chosen as the chunk of data with the
        maximal reconstruction error.

    [Yellin2017] BLOOD CELL DETECTION AND COUNTING IN HOLOGRAPHIC LENS-FREE
    IMAGING BY CONVOLUTIONAL SPARSE DICTIONARY LEARNING AND CODING.
    """
    n_trials, n_channels, n_times = X.shape
    if D.ndim == 2:
        n_times_atom = D.shape[1] - n_channels
    else:
        n_times_atom = D.shape[2]
    patch_rec_error = _patch_reconstruction_error(X, z, D)
    i0 = patch_rec_error.argmax()
    n0, t0 = np.unravel_index(i0, (n_trials, n_times))

    d0 = X[n0, :, t0:t0 + n_times_atom][None]

    if window:
        d0 = d0 * tukey_window(n_times_atom)[None, :]

    if D.ndim == 2:
        return get_uv(d0)

    if D.ndim == 2:
        d0 = prox_uv(d0, uv_constraint=uv_constraint, n_channels=n_channels)
    else:
        d0 = prox_d(d0)

    return d0



#! /usr/bin/env python
import numpy as np
from Cython.Build import cythonize
from setuptools import setup, Extension, find_packages

descr = """Convolutional dictionary learning for noisy signals"""

DISTNAME = 'alphacsc'
DESCRIPTION = descr
MAINTAINER = 'Mainak Jas'
MAINTAINER_EMAIL = 'mainakjas@gmail.com'
LICENSE = 'BSD (3-clause)'
DOWNLOAD_URL = 'https://github.com/alphacsc/alphacsc.git'
VERSION = '0.2.2'

# Add cython extensions
kmc2 = Extension('alphacsc.other.kmc2.kmc2',
                 sources=['alphacsc/other/kmc2/kmc2.pyx'],
                 extra_compile_args=['-O3'])
sdtw = Extension('alphacsc.other.sdtw.soft_dtw_fast',
                 sources=['alphacsc/other/sdtw/soft_dtw_fast.pyx'])
modules = [kmc2, sdtw]

# Create the alphacsc.cython modules
other_modules = [
    "compute_ztX",
    "compute_ztz",
    "coordinate_descent",
    "sparse_conv",
]
for m in other_modules:
    modules.append(
        Extension("alphacsc.cython_code.{}".format(m),
                  sources=["alphacsc/cython_code/{}.pyx".format(m)]))
ext_modules = cythonize(modules)

if __name__ == "__main__":
    setup(
        name=DISTNAME,
        maintainer=MAINTAINER,
        maintainer_email=MAINTAINER_EMAIL,
        description=DESCRIPTION,
        license=LICENSE,
        version=VERSION,
        download_url=DOWNLOAD_URL,
        long_description=open('README.rst').read(),
        classifiers=[
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
            'License :: OSI Approved',
            'Programming Language :: Python',
            'Topic :: Software Development',
            'Topic :: Scientific/Engineering',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX',
            'Operating System :: Unix',
            'Operating System :: MacOS',
        ],
        platforms='any',
        ext_modules=ext_modules,
        packages=find_packages(exclude=["tests"]),
        setup_requires=['Cython', 'numpy'],
        install_requires=[
            'mne',
            'numba',
            'numpy',
            'scipy',
            'joblib',
            'matplotlib',
            'scikit-learn',
        ],
        include_dirs=[np.get_include()], )



# Author: Mathieu Blondel
# License: Simplified BSD

import numpy as np

from .soft_dtw_fast import _soft_dtw
from .soft_dtw_fast import _soft_dtw_grad


class SoftDTW(object):

    def __init__(self, D, gamma=1.0, sakoe_chiba_band=-1):
        """
        Parameters
        ----------
        D: array, shape = [m, n] or distance object
            Distance matrix between elements of two time series.

        gamma: float
            Regularization parameter.
            Lower is less smoothed (closer to true DTW).

        sakoe_chiba_band: int
            If non-negative, the DTW is restricted to a Sakoe-Chiba band around
            the diagonal. The band has a width of 2 * sakoe_chiba_band + 1.

        Attributes
        ----------
        self.R_: array, shape = [m + 2, n + 2]
            Accumulated cost matrix (stored after calling `compute`).
        """
        if hasattr(D, "compute"):
            self.D = D.compute()
        else:
            self.D = D

        self.D = self.D.astype(np.float64)

        self.gamma = gamma
        self.sakoe_chiba_band = sakoe_chiba_band

        if sakoe_chiba_band >= 0:
            assert self.D.shape[0] == self.D.shape[1]

    def compute(self):
        """
        Compute soft-DTW by dynamic programming.

        Returns
        -------
        sdtw: float
            soft-DTW discrepancy.
        """
        m, n = self.D.shape

        # Allocate memory.
        # We need +2 because we use indices starting from 1
        # and to deal with edge cases in the backward recursion.
        self.R_ = np.zeros((m + 2, n + 2), dtype=np.float64)

        _soft_dtw(self.D, self.R_, gamma=self.gamma,
                  sakoe_chiba_band=self.sakoe_chiba_band)

        return self.R_[m, n]

    def grad(self):
        """
        Compute gradient of soft-DTW w.r.t. D by dynamic programming.

        Returns
        -------
        grad: array, shape = [m, n]
            Gradient w.r.t. D.
        """
        if not hasattr(self, "R_"):
            raise ValueError("Needs to call compute() first.")

        m, n = self.D.shape

        # Add an extra row and an extra column to D.
        # Needed to deal with edge cases in the recursion.
        D = np.vstack((self.D, np.zeros(n)))
        D = np.hstack((D, np.zeros((m + 1, 1))))

        # Allocate memory.
        # We need +2 because we use indices starting from 1
        # and to deal with edge cases in the recursion.
        E = np.zeros((m + 2, n + 2))

        _soft_dtw_grad(D, self.R_, E, gamma=self.gamma,
                       sakoe_chiba_band=self.sakoe_chiba_band)

        return E[1:-1, 1:-1]



import itertools
import copy as cp
import os
import json
import datetime

import mne
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scipy import sparse

from .tools import get_calling_script, positive_hash

COLORS = ["#4C72B0", "#55A868", "#C44E52", "#8172B2", "#CCB974", "#64B5CD"]

DEFAULT_CB = {
    'atom': {},
    'z_hat': {},
    'pobj': {}
}
PRINT_KWARGS = {
    'n_atoms': 'K = {}',
    'n_times_atom': 'L={}',
    'reg': r'$\lambda$={}',
    'solver_z': '{}'
}


mpl.rc('mathtext', fontset='cm')


def kde_sklearn(x, x_grid, bandwidth):
    """Kernel Density Estimation with Scikit-learn"""
    n_samples, = x.shape
    if n_samples == 0:
        return np.zeros_like(x_grid)

    window = np.blackman(bandwidth) * bandwidth
    return np.convolve(x, window, 'same')


def plot_activations_density(z_hat, n_times_atom, sfreq=1., threshold=0.01,
                             bandwidth='auto', axes=None, t_min=0,
                             plot_activations=False, colors=None):
    """
    Parameters
    ----------
    z_hat : array, shape (n_atoms, n_trials, n_times_valid)
        The sparse activation matrix.
    n_times_atom : int
        The support of the atom.
    sfreq : float
        Sampling frequency
    threshold : float
        Remove activations (normalized with the max) below this threshold
    bandwidth : float, array of float, or 'auto'
        Bandwidth (in sec) of the kernel
    axes : array of axes, or None
        Axes to plot into
    t_min : float
        Time offset for the xlabel display
    plot_activations : boolean
        If True, the significant activations are plotted as black dots
    colors : list of matplotlib compatible colors
        Colors of the plots
    """
    if sparse.isspmatrix_lil(z_hat[0]):
        z_hat = np.array([z.toarray() for z in z_hat])

    n_atoms, n_trials, n_times_valid = z_hat.shape

    # sum activations over all trials
    z_hat_sum = z_hat.sum(axis=0)

    if bandwidth == 'auto':
        bandwidth = n_times_atom

    if axes is None:
        fig, axes = plt.subplots(n_atoms, num='density',
                                 figsize=(8, 2 + n_atoms * 3))
    axes = np.atleast_1d(axes)

    if colors is None:
        colors = itertools.cycle(COLORS)
    for ax, activations, color in zip(axes.ravel(), z_hat_sum, colors):
        ax.clear()
        time_instants = np.arange(n_times_valid) / float(sfreq) + t_min
        selection = activations > threshold * z_hat_sum.max()
        n_elements = selection.sum()

        if n_elements == 0:
            ax.plot(time_instants, np.zeros_like(time_instants))
            continue

        # plot the activations as black dots
        if plot_activations:
            ax.plot(time_instants[selection],
                    activations[selection] / activations[selection].max(), '.',
                    color='k')

        window = np.blackman(bandwidth)
        smooth_activations = np.convolve(activations, window, 'same')
        ax.fill_between(time_instants, smooth_activations, color=color,
                        alpha=0.5)

    return axes


def try_plot_activations_density():
    """Just try it"""
    z_hat = np.random.randn(3, 1, 1000)
    z_hat[z_hat < 2] = 0
    z_hat[:, :, :500] /= 6.

    plot_activations_density(z_hat, 150, plot_activations=True)
    plt.show()


def plot_data(X, plot_types=None):
    """Plot the data.

    Parameters
    ----------
    X : list
        A list of arrays of shape (n_trials, n_times).
        E.g., one could give [X, X_hat]
    plot_types : list of str
        If None, plt.plot for all.
        E.g., plot_data([X, z], ['plot', 'stem'])
    """
    import matplotlib.pyplot as plt

    if not isinstance(X, list):
        raise ValueError('Got %s. It must be a list' % type(X))

    if plot_types is None:
        plot_types = ['plot' for ii in range(len(X))]

    if not isinstance(plot_types, list):
        raise ValueError('Got %s. It must be a list' % type(plot_types))
    if len(plot_types) != len(X):
        raise ValueError('X and plot_types must be of same length')

    def _onclick(event):
        orig_ax = event.inaxes
        fig, ax = plt.subplots(1)
        ax.set_axis_bgcolor('white')
        for jj in range(len(plot_types)):
            if orig_ax._plot_types[jj] == 'plot':
                ax.plot(orig_ax._X[jj])
            elif orig_ax._plot_types[jj] == 'stem':
                ax.plot(orig_ax._X[jj], '-o')
        plt.title('%s' % orig_ax._name)
        plt.show()

    n_trials = X[0].shape[0]
    fig, axes = plt.subplots(n_trials, 1, sharex=True, sharey=True)
    fig.canvas.mpl_connect('button_press_event', _onclick)
    fig.patch.set_facecolor('white')
    for ii in range(n_trials):
        for jj in range(len(X)):
            if plot_types[jj] == 'plot':
                axes[ii].plot(X[jj][ii])
            elif plot_types[jj] == 'stem':
                axes[ii].plot(X[jj][ii], '-o')
        axes[ii].get_yaxis().set_ticks([])
        axes[ii].set_ylabel('Trial %d' % (ii + 1), rotation=0, ha='right')
        axes[ii]._name = 'Trial %d' % (ii + 1)
        axes[ii]._plot_types = plot_types
        axes[ii]._X = [X[jj][ii] for jj in range(len(X))]
    plt.xlabel('Time')
    plt.show()


def plot_callback(X, info, n_atoms, layout=None):
    n_trials, n_channels, n_times = X.shape

    n_atoms_plot = min(15, n_atoms)

    fig, axes = plt.subplots(nrows=n_atoms, num='atoms', figsize=(10, 8))
    fig_z, axes_z = plt.subplots(nrows=n_atoms, num='z', figsize=(10, 8),
                                 sharex=True, sharey=True)
    fig_topo, axes_topo = plt.subplots(1, n_atoms_plot, figsize=(12, 3))

    if n_atoms == 1:
        axes_topo, axes = [axes_topo], [axes]
        if n_trials == 1:
            axes_z = [axes_z]

    if layout is None:
        layout = mne.channels.find_layout(info)

    def callback(X, uv_hat, z_hat, reg):
        n_times_valid = z_hat.shape[-1]
        n_times_atom = uv_hat.shape[1] - n_channels
        times_z = np.arange(n_times_valid) / info['sfreq']
        times_v = np.arange(n_times_atom) / info['sfreq']

        this_info = cp.deepcopy(info)
        this_info['sfreq'] = 1.
        patterns = mne.EvokedArray(uv_hat[:n_atoms_plot, :n_channels].T,
                                   this_info, tmin=0)
        patterns.plot_topomap(times=np.arange(n_atoms_plot),
                              layout=layout, axes=axes_topo, time_unit='s',
                              time_format='Atom%01d', show=False)

        if axes[0].lines == []:
            for k in range(n_atoms):
                axes[k].plot(times_v, uv_hat[k, n_channels:].T)
                axes[k].grid(True)
        else:
            for ax, uv in zip(axes, uv_hat):
                ax.lines[0].set_ydata(uv[n_channels:])
                ax.relim()  # make sure all the data fits
                ax.autoscale_view(True, True, True)
        if n_trials == 1:
            if axes_z[0].lines == []:
                for k in range(n_atoms):
                    axes_z[k].plot(times_z, z_hat[0, k])
                    axes_z[k].grid(True)
            else:
                for ax, z in zip(axes_z, z_hat[0, :]):
                    ax.lines[0].set_ydata(z)
                    ax.relim()  # make sure all the data fits
                    ax.autoscale_view(True, True, True)
        else:
            extent = [times_z[0], times_z[-1], 1, len(z_hat[0])]
            for k in range(n_atoms):
                im = axes_z[k].imshow(z_hat[:, k], cmap='hot', origin='lower',
                                      extent=extent,
                                      clim=(0.0, z_hat.max()), aspect='auto')
            fig_z.subplots_adjust(right=0.8)
            cbar_ax = fig_z.add_axes([0.86, 0.10, 0.03, 0.8])
            fig_z.colorbar(im, ax=None, cax=cbar_ax)

        fig.canvas.draw()
        fig_topo.canvas.draw()
        fig_z.canvas.draw()
        plt.pause(.001)

    return callback


def plot_or_replot(data, axes=None, sfreq=1):
    """Given a list of axes and the ydata of each axis, cleanly update the plot

    Parameters
    ----------
    data : list of ydata
        Data to plot, or to update the axes lines.
    axes : list of Axes or None
        Axes to update.
    sfreq : float or None
        Value to compute the xlabel.
    """
    if axes is None:
        K = len(data)
        n_rows, n_cols = min(K, 5), max(1, K // 5)
        fig, axes = plt.subplots(n_rows, n_cols, squeeze=False)
        axes = axes.ravel()
    if axes[0].lines == []:
        color_cycle = itertools.cycle(COLORS)
        for ax, xk, color in zip(axes, data, color_cycle):
            ax.plot(np.arange(xk.shape[-1]) / sfreq, xk, c=color)
            ax.grid(True)
    else:
        for ax, xk in zip(axes, data):
            ax.lines[0].set_xdata(np.arange(xk.shape[-1]) / sfreq)
            ax.lines[0].set_ydata(xk)
            ax.relim()  # make sure all the data fits
            ax.autoscale_view(True, True, True)


def get_callback_csc(csc_kwargs, config=DEFAULT_CB, info={}):
    """Setup and return a callback for csc scripts

    Parameters
    ----------
    csc_kwargs : dict
        Parameters used to run the csc experiment. It will be hashed and used
        in the name of the files.
    info : dict
        Information on the signal.
    config : dict
        The key should be in {'atom', 'Xhat', 'z_hat', 'topo'}. and the value
        should be a configuration dictionary. The considered config are:
            'share': If True, share X and Y axis for all the axes.
            'info': Required for 'topo', should be the epoch.info.

    Return
    ------
    callback : callable
        A callback function that can be used to periodically plot and save
        the csc results.
    """
    csc_hash = positive_hash(json.dumps(csc_kwargs, sort_keys=True))
    csc_time = datetime.datetime.now().strftime('%d_%m_%y_%Hh%M')
    csc_script = get_calling_script()

    csc_name = "{}_csc_{}".format(csc_time, csc_hash)

    if not os.path.exists('save_exp'):
        os.mkdir('save_exp')
    save_dir = 'save_exp/{}'.format(csc_script)
    if not os.path.exists(save_dir):
        os.mkdir(save_dir)

    # Dump the arguments in a json to easily find introspect them
    with open('{}/{}.json'.format(save_dir, csc_name), 'w') as f:
        json.dump(csc_kwargs, f, sort_keys=True)

    n_atoms = csc_kwargs.get('n_atoms')
    n_times_atom = csc_kwargs.get('n_times_atom')

    # Extract plot information
    sfreq = info.get('sfreq', 1)
    t_min = info.get('t_min', 0)

    # Compute the shape of the subplot
    ncols = int(np.ceil(n_atoms / 5.))
    nrows = min(5, n_atoms)

    # Clean previous figures
    plt.close('all')

    figs = {}
    for name, conf in config.items():
        assert name in ['atom', 'z_hat', 'Xhat', 'topo', 'pobj']

        fname = "{} - {}".format(name, os.getpid())
        shared_axes = conf.get('share', True)
        width = 5
        nc, nr = ncols, nrows
        if name == 'topo':
            width = 3
        if name == 'pobj':
            nc = nr = 1
            width = 10
        f, axes = plt.subplots(nrows=nr, ncols=nc, num=fname,
                               figsize=(width * nc, 10), squeeze=False,
                               sharex=shared_axes, sharey=shared_axes)

        text = [PRINT_KWARGS.get(k, "{}={{}}".format(k)).format(v)
                for k, v in csc_kwargs.items() if k in PRINT_KWARGS]
        text = "\t".join(text)
        f.suptitle(text, fontsize=18, wrap=True)

        if name == 'pobj':
            axes[0, 0].set_xscale('log')
            axes[0, 0].set_yscale('log')

        figs[name] = (f, axes)
    for f, _ in figs.values():
        f.tight_layout()
        f.subplots_adjust(top=.9)
        f.canvas.draw()
    plt.pause(.1)

    def callback(X, uv_hat, z_hat, pobj):

        n_channels = X.shape[1]

        if 'topo' in figs:
            axes = figs['topo'][1]
            for idx, ax in enumerate(axes.ravel()):
                ax.clear()
                mne.viz.plot_topomap(uv_hat[idx, :n_channels], info,
                                     axes=ax, show=False)
                ax.set_title('atom %d' % idx)

        if 'atom' in figs:
            axes = figs['atom'][1].ravel()
            if csc_kwargs.get('rank1', True):
                plot_or_replot(uv_hat[:, n_channels:], axes, sfreq=sfreq)
            else:
                i0 = uv_hat.std(axis=-1).argmax(axis=-1)
                plot_or_replot(
                    [uv_hat[k, i0[k]] for k in range(n_atoms)], axes)

        if 'Xhat' in figs:
            X_hat = [np.convolve(z_k, uvk[n_channels:])
                     for z_k, uvk in zip(z_hat[0], uv_hat)]
            axes = figs['Xhat'][1].ravel()
            plot_or_replot(X_hat, axes)

        if 'z_hat' in figs:
            axes = figs['z_hat'][1].ravel()
            plot_activations_density(z_hat, n_times_atom, sfreq=sfreq,
                                     plot_activations=False, axes=axes,
                                     bandwidth='auto', t_min=t_min)

        if 'pobj' in figs:
            axes = figs['pobj'][1].ravel()
            cost = np.array([pobj])
            plot_or_replot(cost - np.min(cost) + 1e-6, axes)

        # Update and save the figures
        for fig_name, (f, _) in figs.items():
            f.canvas.draw()
            f.savefig('{}/{}_{}.png'.format(save_dir, csc_name, fig_name))
        plt.pause(.001)

    return callback



import os
import mne
import numpy as np
from joblib import Memory
from scipy.signal import tukey

try:
    from ..utils import check_random_state
except ValueError:
    from alphacsc.utils import check_random_state


mem = Memory(location='.', verbose=0)


@mem.cache()
def load_data(n_trials=40, n_channels=1, n_times=6, sigma=.05, sfreq=300,
              f_noise=True, random_state=None, n_jobs=4):
    """Simulate data with 10Hz mu-wave and 10Hz oscillations.

    Parameters
    ----------
    n_trials : int
        Number of simulated signals
    n_channels : int
        Number of channels in the signals
    n_times : float
        Length of the generated signals, in seconds. The generated signal
        will have length n_times * sfreq
    sigma : float
        Additive noise level in the signal
    sfreq : int
        Sampling rate for the signal
    f_noise : boolean
        If set to True, use MNE empty room data as a noise in the signal
    n_jobs : int
        Number of processes used to filter the f_noise
    random_state : int | None
        State to seed the random number generator

    Return
    ------
    X : ndarray, shape (n_trials, n_channels, n_times)
        Simulated 10Hz sinusoidal signals with 10Hz mu-wave between 2sec and
        [3, 3.5]sec. some random phases are applied to the different part of
        the signal.
    info : dict
        Contains the topomap 'u' associated to each component of the signal.
    """
    rng = check_random_state(random_state)

    freq = 10  # Generate 10Hz mu-wave
    phase_shift = rng.rand(n_trials, 1) * sfreq * np.pi
    t0 = 2
    L = 1. + .5 * rng.rand(n_trials, 1)
    t = (np.linspace(0., n_times, int(n_times * sfreq)))
    mask = (t[None] > t0) * (t[None] < t0 + L)
    t *= 2 * np.pi * freq
    t = t[None] + phase_shift
    noisy_phase = .5 * np.sin(t / (3 * np.sqrt(2)))
    phi_t = t + noisy_phase
    signal = np.sin(phi_t + np.cos(phi_t) * mask)

    U = rng.randn(1, n_channels, 1)
    U_mu = rng.randn(1, n_channels, 1)
    U /= np.sqrt((U * U).sum()) * np.sign(U.sum())
    U_mu /= np.sqrt((U_mu * U_mu).sum()) * np.sign(U_mu.sum())
    signal = (U + (U_mu - U) * mask[:, None]) * signal[:, None]

    # signal += sigma * rng.randn(*signal.shape)

    # generate noise from empty room noise inthe mne.sample dataset
    if f_noise:
        data_path = os.path.join(mne.datasets.sample.data_path(), 'MEG',
                                 'sample')
        raw = mne.io.read_raw_fif(os.path.join(data_path, 'ernoise_raw.fif'),
                                  preload=True)

        raw.pick_types(meg='mag')
        nyquist = raw.info['sfreq'] / 2.
        raw.notch_filter(np.arange(60, nyquist - 10., 60), n_jobs=n_jobs)
        raw.filter(.5, None, n_jobs=n_jobs)
        X = raw[:][0]
        X /= X.std(axis=1, keepdims=True)
        max_channels, T_max = X.shape

        channels = rng.choice(max_channels, n_channels)
        L_sig = int(n_times * sfreq)
        for i in range(n_trials):
            t = rng.choice(T_max - L_sig)
            signal[i] += sigma * X[channels, t:t + L_sig]
    else:
        signal += sigma * rng.randn(signal.shape)

    signal *= tukey(signal.shape[-1], alpha=0.05)[None, None, :]

    info = {}
    info['u'] = np.r_[U[:, :, 0], U_mu[:, :, 0]]

    return signal, info


if __name__ == "__main__":

    import matplotlib.pyplot as plt

    load_data.clear()
    X, info = load_data(n_trials=1000, sigma=1, n_channels=1)
    plt.plot(np.mean(X, axis=0).T)
    plt.plot(X[0].T)
    plt.show()
